//Chris Hunter
//chunter@rice.edu
/*****************CSMA*****************
The csmaMac is a modified ALOHA MAC that
serves as an example for novel MAC
development. Nodes transmit whenever
they have information to transmit, and only
move on to the next packet once the original
transmit is acknowledged (ACKed). If no ACK
is received, a collision is inferred and the
packet is re-transmitted.

By default, the MAC also implements carrier-
sensing multiple-access with collision-
avoidance (CSMA-CA). This functionality is
built into hardware peripherals in the project
so very little software state is affected.

In its current state, the project acts as
a true ethernet MAC-level wireless bridge.
Any ethernet activity that appears on one
WARP will be sent to another via the custom
wireless link.

Additionally, there is a long packet mode
to test link quality. The MAC itself will
generate "dummy packets" and send them to
another node. Link quality can be oberved
on the LEDs without intervention from a
computer.
*****************************************/

#include "warpmac.h"
#include "xmk.h"
#include "xparameters.h"
#include <stdio.h>
#include <sys/bufmalloc.h>
#include <pthread.h>
#include <errno.h>
#include <stdlib.h>
#include "xutil.h"
#include "xcache_l.h"
#include "xexception_l.h"
#include "xcache_l.h"
#include "xutil.h"


#define MULTIHOPMODE

//Define a routing table definition
typedef struct {
	unsigned char addr[6];
} route;
unsigned char myAddr[6];
route routeTable[16];
unsigned char myID;


Macframe txBuffer; //Buffer for holding a packet-to-xmit across multiple retransmissions
				   //this is a global variable to enable easy-access from callbacks,
				   //interrupt handlers, etc.

Macframe rxBuffer;
unsigned char relayCurrTx = 0;

//MAC Revision
#define ACKMAC 1

//Packet Types
#define DATAPACKET 0
#define ACKPACKET 1

int uphandle() {
	warpmac_setTargetPlus();
}

int middlehandle() {
	warpmac_setTargetMinus();
}

int timerExpire(unsigned char timerType){
	/*This function is responsible for handling TIMEOUTs and BACKOFFs.
	It is registered using the warpmac_setTimerHandler function in
	myMac_main. The job responsibilities of this function are to:
	-increase the contention window upon the expiration of a TIMEOUT
	-initiate a BACKOFF timer upon the expiration of a TIMEOUT
	-retransmit a packet upon the expiration of a BACKOFF*/
	int status;

	switch(timerType){
		case TIMEOUT:
			if(txBuffer.isNew){
				status = warpmac_incrementResend(&txBuffer);
				if(status == 0){
					if(myID==2) warpmac_setRxBuffer(&txBuffer);
					return 0;
				}
				warpmac_setTimer(BACKOFF);
				return 0;
			}
				break;
		case BACKOFF:

	if(warpmac_carrierSense()){ //This is somewhat of an overkill
		warpmac_sendOfdm(&txBuffer);
		warpmac_setTimer(TIMEOUT);
	}
	else{
		warpmac_setTimer(BACKOFF);
	}
	break;
	return 0;
	}
}

int* ethernet_callback(Xuint8 * frame, Xuint32 length){
	/*This function is called by the ethernet MAC drivers
	when a packet is available to send. This function fills
	the Macframe transmit buffer with the packet and sends
	it over the OFDM link*/

	//warpmac_allocatePayload(&txBuffer,length);
	txBuffer.currReSend = 0;
	txBuffer.isNew = 1;
	txBuffer.pktRev = ACKMAC;
	txBuffer.length = length;
	txBuffer.pktType = DATAPACKET;
	memcpy(txBuffer.srcAddr,myAddr,6); //Copy MAC address into source field


	#ifndef MULTIHOPMODE
		memcpy(txBuffer.destAddr,routeTable[(myID+1)%2].addr,6);
	#endif

	#ifdef MULTIHOPMODE
		memcpy(txBuffer.destAddr,routeTable[2].addr,6);
	#endif


	memcpy(txBuffer.data, frame, length);




	if(warpmac_carrierSense()){

		warpmac_sendOfdm(&txBuffer);
		warpmac_setTimer(TIMEOUT);
	}
	else{
		warpmac_setTimer(BACKOFF);
	}
	return 0;
}


int receiveBadPacket(Macframe* packet) {

}

int receiveGoodPacket(Macframe* packet) {
	/*This function processes the received packet to see if it
	was addressed to this node, and sends an ACK if it is a
	data packet. Also, it pushes the received packet out to
	ethernet.*/

	//warpmac_incrementLEDHigh();
	//IMPORTANT NOTE: Even though this function is passed a Macframe structure,
	//the data field is not yet filled. The user's MAC should process the packet
	//based on the header information, send out an ACK if necessary, and then
	//copy the payload out of the PHY with the warpmac_writePacket command.
	//if(warpmac_addressedToThem(packet,routeTable[0].addr)) warpmac_incrementLEDHigh();
	//if(warpmac_addressedToThem(packet,routeTable[2].addr)) warpmac_incrementLEDLow();

//warpmac_printDestAddr(packet);
warpmac_incrementLEDHigh();

	if(warpmac_addressedToMe(packet)){
	warpmac_incrementLEDLow();
	Macframe ackPacket;
		switch(packet->pktType){
			case ACKPACKET:
				if(warpmac_inTimeout()){
					warpmac_clearTimer(TIMEOUT);
					//warpmac_freePayload(&txBuffer);
					txBuffer.currReSend = 0;
					txBuffer.isNew = 0;
					//Delay is necessary to give receiver time to push packet over ethernet...
					//this will be removed when DMA is implemented

					if(myID!=2){
						usleep(125);//125
					}
					else{
						warpmac_setRxBuffer(&txBuffer);
					}
				}
				break;
			case DATAPACKET:

				ackPacket.pktRev = ACKMAC;
				ackPacket.length = 0;
				ackPacket.pktType = ACKPACKET;
				memcpy(ackPacket.srcAddr,myAddr,6);
				memcpy(ackPacket.destAddr,packet->srcAddr,6);

				warpmac_sendOfdm(&ackPacket);


				if(myID!=2){
					//warpmac_allocatePayload(packet,packet->length);
					packet->currReSend = 0;
					packet->isNew = 1;
					warpmac_copyPayload(packet);
					warpmac_sendEthernet(packet);
					//warpmac_freePayload(packet); //Fill the rx queue with the received packet
					packet->currReSend = 0;
					packet->isNew = 0;
				}
				else{
				//****
					if(txBuffer.isNew==0){
									//warpmac_allocatePayload(packet,packet->length);
						txBuffer.currReSend = 0;
						txBuffer.isNew = 1;
					warpmac_copyPayload(packet);
					//warpmac_allocatePayload(&txBuffer,packet->length);
					//txBuffer.pktRev = ACKMAC;
					//txBuffer.length = packet->length;
					//txBuffer.pktType = DATAPACKET;


					unsigned char destAddr[6];

					int i;
					for(i=0;i<2;i++){
						if(warpmac_addressedFromThem(packet,routeTable[i].addr)){

								warpmac_leftHex((unsigned char)((i+1)%2));
								//warpmac_leftHex(9);
								memcpy(txBuffer.destAddr,routeTable[(i+1)%2].addr,6);
						}
					}
					//memcpy(txBuffer.destAddr,routeTable[0].addr,6);

					memcpy(txBuffer.srcAddr,myAddr,6); //Copy MAC address into source field
					//memcpy(txBuffer.data, packet->data, packet->length);


					//if(warpmac_carrierSense()){

						warpmac_sendOfdm(&txBuffer);
						warpmac_setTimer(TIMEOUT);
					//}
					//else{
					//	warpmac_setTimer(BACKOFF);
					//}
					warpmac_setRxBuffer(&rxBuffer);
					}
					//****
				}





				//warpmac_sendEthernet(packet);
				//warpmac_freePayload(packet);
				break;
		}
	}
	return 0;
}

/*MYMAC_MAIN- function is instantiating the MAC framework and looping in an idle state*/
void* myMac_main(){

	//Read Dip Switch value from FPGA board.
	//This value will be used as an index into the routing table for other nodes
	myID = warpmac_getMyId();

	//Create an arbitrary address for this node
	unsigned char tmpAddr[6] = {0x16,0x24,0x63,0x53,0xe2,0xc2+myID};
	memcpy(myAddr,tmpAddr,6);

	//Fill an arbitrary routing table so that nodes know each others' addresses
	unsigned char i;
	for(i=0;i<16;i++){
		routeTable[i].addr[0] = myAddr[0];
		routeTable[i].addr[1] = myAddr[1];
		routeTable[i].addr[2] = myAddr[2];
		routeTable[i].addr[3] = myAddr[3];
		routeTable[i].addr[4] = myAddr[4];
		routeTable[i].addr[5] = myAddr[5]+i-myID;
	}


	//Initialize the framework
	warpmac_init();
	warpmac_setMacAddr(&myAddr);
	warpmac_setMaxResend(4);
	warpmac_setMaxCW(4);
	warpmac_setTimeout(400);
	warpmac_setSlotTime(9);
	warpmac_disableSequencing();




	if(myID!=2){
		warpmac_setRxBuffer(&rxBuffer);
	}

	else{
		warpmac_setRxBuffer(&txBuffer);
	}


	warpmac_allocatePayload(&txBuffer,(Xuint32)2000);
	warpmac_allocatePayload(&rxBuffer,(Xuint32)2000);


xil_printf("Tx Payload Addr 0x%x\r\n",txBuffer.data);
xil_printf("Rx Payload Addr 0x%x\r\n",rxBuffer.data);

	warpmac_setGoodPacketHandler(receiveGoodPacket);
	warpmac_setBadPacketHandler(receiveBadPacket);

	warpmac_setUpButtonHandler(uphandle);
	warpmac_setMiddleButtonHandler(middlehandle);


	warpmac_setTimerHandler(timerExpire);
	warpmac_setChannel(11);
	warpmac_enableSisoMode();
	warpmac_enableCSMA();

	while(1){
			if(txBuffer.isNew==0){
				if(myID!=2){
					warpmac_pollEthernet(ethernet_callback);
				}
			}
	}
	pthread_exit (NULL);
}

int main(){
	xilkernel_main();
}



