//Chris Hunter
//WARPMAC Framework
/*****************WARPMAC*****************
Description: This framework allows for custom
MAC implementations on WARP. For a description
of the functions, see:
http://warp.rice.edu/trac/wiki/WARPMAC
******************************************/
#include "xparameters.h"
#include "xmk.h"
#include <sys/bufmalloc.h>
#include "xemac_l.h"
#include "xemac.h"
#include <os_config.h>
#include <xstatus.h>
#include <sys/intr.h>
#include "xintc.h"
#include "xgpio.h"
#include <errno.h>

//Header files for WARP peripheral drivers
#include "ofdm_TxRx_mimo.h"
#include "ofdm_AGC_mimo.h"
#include "ofdm_timer.h"
#include "ofdm_pktDetector_mimo_gateways.h"
#include "EEPROM.h"
#include "radio_controller.h"

//Header files for WARP libraries
#include "warpmac.h"

Maccontrol controlStruct;
XGpio Gpio;

////Globals for PHY Debugging////
int agctarget = -15;
int agcnoiseEst = -95;
unsigned int sisoMode = 0;
unsigned int numTrainingSyms = 4;
unsigned int numBaseRate = 2;
#define INIT_KPVAL 0x2C3C38//0xA6800
#define INIT_KIVAL 0x960//0xDE00
unsigned int KPval = INIT_KPVAL;
unsigned int KIval = INIT_KIVAL;
unsigned int pktDly = 58;//251;
unsigned int pktRawThresh = 4000;
unsigned char globalReset = 0;
/////////////////////////////////

////Xilkernel Parameters////
//This "reserves" a piece of memory for xilkernel to malloc in
static  membuf_t mbuftACK;
char    membufACK[BUFWIDTH][BUFHEIGHT] __attribute__ ((aligned(4)));
pthread_attr_t attr;
pthread_t tid[N_THREADS];
////////////////////////////

///Global LED states////
unsigned int ledStates[LEDLEN] = {1,2,4,8,4,2};
unsigned int ledCounter = 0;
unsigned char ledStatesIndex = 0;
unsigned int ledStatesLow[2] = {1,2};
unsigned int ledCounterLow = 0;
unsigned char ledStatesIndexLow = 0;
unsigned int ledStatesHigh[2] = {4,8};
unsigned int ledCounterHigh = 0;
unsigned char ledStatesIndexHigh = 0;
////////////////////////

////Ethernet Parameters////
XEmac Emac;
XEmac *InstancePtr = &Emac;
Xuint32 RXBuffer[MY_XEM_MAX_FRAME_SIZE_IN_WORDS];
Xuint32 TXBuffer[MY_XEM_MAX_FRAME_SIZE_IN_WORDS];
static Xuint8 LocalAddress[XEM_MAC_ADDR_SIZE] = {0x06, 0x05, 0x04, 0x03, 0x02, 0x01}; //wired MAC address... just filler... not used
																					  ///////////////////////////

////User Callback Pointers////
//These the addresses pointed to by these variables are set by the
//user via the callback registration functions
void (*usr_upbutton) ();
void (*usr_leftbutton) ();
void (*usr_middlebutton) ();
void (*usr_rightbutton) ();
void (*usr_badPacketHandler) ();
void (*usr_goodPacketHandler) ();
void (*usr_timerHandler) ();
//////////////////////////////

Macframe* rxPacket;

void nullHandler(void* param){
};

/*WARPMAC_INCREMENTLED - Every BLINKEVERY (in warpmac.h) times this function is called,
the user LEDs will switch state. The end result is a visualization that looks like Kitt
from Knightrider.*/
void warpmac_incrementLED(){
	if(ledCounter==0){
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, ledStates[ledStatesIndex]);
		ledStatesIndex = (ledStatesIndex+1)%LEDLEN;
	}
	ledCounter = (ledCounter+1)%BLINKEVERY;
}

/*WARPMAC_INCREMENTLEDLOW - Much like WARPMAC_INCREMENTLED, but allows for independent
control of the bottom two LEDs*/
void warpmac_incrementLEDLow(){
	if(ledCounterLow==0){
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, (ledStatesLow[ledStatesIndexLow]|ledStatesHigh[ledStatesIndexHigh]));
		ledStatesIndexLow = (ledStatesIndexLow+1)%2;
	}
	ledCounterLow = (ledCounterLow+1)%BLINKEVERY;
}

/*WARPMAC_INCREMENTLEDHIGH - Much like WARPMAC_INCREMENTLED, but allows for independent
control of the top two LEDs*/
void warpmac_incrementLEDHigh(){
	if(ledCounterHigh==0){
		//xil_printf("%x\r\n", (ledStatesLow[ledStatesIndexLow]|ledStatesHigh[ledStatesIndexHigh]));
		XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, (ledStatesLow[ledStatesIndexLow]|ledStatesHigh[ledStatesIndexHigh]));
		ledStatesIndexHigh = (ledStatesIndexHigh+1)%2;

	}
	ledCounterHigh = (ledCounterHigh+1)%BLINKEVERY;
}


/*TIMER_A_INT_HANDLER - interrupt handler for the first timer peripheral
The timer interrupt handler gets called when a hardware timer set by the
user expires. The user's callback is called upon the completion of this event */
void timer_a_int_handler(void * baseaddr_p) {
	/* Clear the timer interrupt */
	ofdm_timer_clearInterrupt();
	int baseaddr = (int) baseaddr_p;
	controlStruct.timerEnabled = 0;
	//Call User Function
	usr_timerHandler(controlStruct.timerType);

}


/*rxPhyBad - interrupt handler for the receive PHY layer
This function is the ISR for the RX physical layer peripheral.
This function is only called if the packet fails CRC. The received
packet is passed to the user's callback.*/
void rxPhyBad(void *baseaddr_p) {

//	unsigned char header[24];
//	unsigned int numPayloadBytes;
//	int length;
//	mimo_ofdmRx_copyHeader(header,24); //Copy just the header
//	rxPacket->pktType = header[13];
//	rxPacket->destAddr[0] = header[3];
//	rxPacket->destAddr[1] = header[2];
//	rxPacket->destAddr[2] = header[1];
//	rxPacket->destAddr[3] = header[0];
//	rxPacket->destAddr[4] = header[7];
//	rxPacket->destAddr[5] = header[6];
//	rxPacket->srcAddr[0] = header[5];
//	rxPacket->srcAddr[1] = header[4];
//	rxPacket->srcAddr[2] = header[11];
//	rxPacket->srcAddr[3] = header[10];
//	rxPacket->srcAddr[4] = header[9];
//	rxPacket->srcAddr[5] = header[8];
//	rxPacket->pktRev = header[15];
//	unsigned int a = (unsigned int)header[18];
//	unsigned int b = (unsigned int)header[19];
//	numPayloadBytes = a*256 + b;
//	rxPacket->length = numPayloadBytes - 24 - EXTRASYMBOLS*12 - 4; //4 bytes for checksum
//	if(rxPacket->length < 0 || rxPacket->length>MY_XEM_MAX_FRAME_SIZE){
//		rxPacket->length=0;
//	}
//
//	//Call User Function
	usr_badPacketHandler(rxPacket);
	mimo_mac2phy_PktAck();
}


/*rxPhyGood - interrupt handler for the receive PHY layer
This function is the ISR for the RX physical layer peripheral.
This function is only called if the packet has no bit errors
(i.e., the packet passes CRCs in the PHY). The function copies
the contents from the PHY into a temporary buffer, which then
gets converted to the custom frame structure that the MAC
code understands. The received packet is passed to the
user's callback.*/
void rxPhyGood(void *baseaddr_p) {


	unsigned char header[24];
	unsigned int numPayloadBytes;
	int length;
	mimo_ofdmRx_copyHeader(header,24); //Copy just the header
	rxPacket->pktType = header[13];
	rxPacket->destAddr[0] = header[3];
	rxPacket->destAddr[1] = header[2];
	rxPacket->destAddr[2] = header[1];
	rxPacket->destAddr[3] = header[0];
	rxPacket->destAddr[4] = header[7];
	rxPacket->destAddr[5] = header[6];
	rxPacket->srcAddr[0] = header[5];
	rxPacket->srcAddr[1] = header[4];
	rxPacket->srcAddr[2] = header[11];
	rxPacket->srcAddr[3] = header[10];
	rxPacket->srcAddr[4] = header[9];
	rxPacket->srcAddr[5] = header[8];
	rxPacket->pktRev = header[15];
	unsigned int a = (unsigned int)header[18];
	unsigned int b = (unsigned int)header[19];


	if(controlStruct.sequenceVec[rxPacket->srcAddr[5]].sequenceNumber[1] == header[16]){
		controlStruct.packetDupe=1;	//if sequencing is enabled, this will be used to keep the packet from
									//being sent over ethernet
	}
	else{
		controlStruct.packetDupe=0;
		controlStruct.sequenceVec[rxPacket->srcAddr[5]].sequenceNumber[1] = header[16];
	}

	numPayloadBytes = a*256 + b;//(*(unsigned short *)(0xFF101000+18));
		rxPacket->length = numPayloadBytes - 24 - EXTRASYMBOLS*12 - 4; //4 bytes for checksum
		if(rxPacket->length < 0 || rxPacket->length>MY_XEM_MAX_FRAME_SIZE){
			rxPacket->length=0;
		}
		usr_goodPacketHandler(rxPacket);
		mimo_mac2phy_PktAck();
}

/*PB_INT_HANDLER - interrupt handler for the pushbuttons on the WARP board. This is a low-priority
interrupt because it is primarily used for debugging purposes. The various user callbacks are
executed, depending upon which button was depressed.*/
void pb_int_handler(void *baseaddr_p) {

	Xuint32 dsr;
	XStatus stat;
	char lock;

	dsr = XGpio_mGetDataReg(XPAR_PUSH_BUTTONS_4BIT_BASEADDR, 2);
	switch(dsr) {

		case 0x01:
			usr_middlebutton();
			break;

		case 0x02:
			usr_rightbutton();
			break;

		case 0x08:
			usr_upbutton();
			break;
		case 0x04:
			usr_leftbutton();
			break;

		default : {
		}

	}
	XGpio_InterruptClear(&Gpio, 0x3); // clear the interrupt
}

/*WARPMAC_CLEARTIMER - This function stops the ofdm_timer. Additionally it
will return the amount of time remaining before expiration.*/
int warpmac_clearTimer(){
	int time;
	time = ofdm_timer_timeLeft();
	ofdm_timer_stop();
	return time;
}

/*WARPMAC_SETTIMERVAL - This function sets the ofdm_timer to countdown for
a given number of clock cycles.*/
void warpmac_setTimerVal(Xuint32 clocks){
	ofdm_timer_setVal(clocks);
}

/*WARPMAC_STARTTIMER - This function starts the ofdm_timer in either
a CSMA or non-CSMA mode.*/
void warpmac_startTimer(unsigned char mode){
	ofdm_timer_setMode(mode);
	ofdm_timer_start();
}

/*RANDNUM- generates a uniform random value between [0,N-1]*/
int randNum(int N){
	int ans;
	int temp;
	temp = rand();
	ans = temp>>(32-(N+6));
	return ans;
}

/*WARPMAC_CARRIERSENSE- returns a value corresponding to the instantaneous channel condition.*/
unsigned int warpmac_carrierSense(){
	return *((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_idleDifs_OFFSET));
}

/*SETTIMER- function is responsible for enabling the hardware timer peripherals*/
int warpmac_setTimer(int type){
	unsigned int setTime;
	int myRandNum;
	int retval;
	switch(type) {
		case TIMEOUT:
			controlStruct.timerType=TIMEOUT;
			setTime = controlStruct.timeout*50;
			warpmac_setTimerVal(setTime);
			warpmac_startTimer(DISABLECSMA);
			return 0;
			break;

		case BACKOFF:
			controlStruct.timerType=BACKOFF;
			myRandNum = randNum(controlStruct.currBackoff);
			setTime = myRandNum*controlStruct.slotTime*50;
			if(controlStruct.currBackoff < controlStruct.maxBackoff){
				/*Note: Currently, I never decrement a backoff... I only reset it upon the next frame. Technically, CSMA-CA
				should decrement the window under certain circumstances (check the 802.11 standard for details)*/
				controlStruct.currBackoff = controlStruct.currBackoff + 1;
			}

				/* Set the reset value of the timer (seconds * 50,000,000) */
				if(setTime != 0){
					warpmac_setTimerVal(setTime);
					warpmac_startTimer(ENABLECSMA);
					return 0;
				}

				if(setTime == 0){
					usr_timerHandler(controlStruct.timerType);
				}
				break;

	}
}


/*WARPMAC_INIT- initializes the framework and all hardware peripherals.*/
int warpmac_init() {

	usr_upbutton = nullHandler;
	usr_leftbutton = nullHandler;
	usr_middlebutton = nullHandler;
	usr_rightbutton = nullHandler;
	usr_badPacketHandler = nullHandler;
	usr_goodPacketHandler = nullHandler;



	xil_printf("Initializing WARPMAC...\r\n");

	xil_printf("	Initializing Ethernet...");
	/////////////////////////ETHERNET///////////////////////
	int i,l;
	XStatus Status;
	Xuint32 Options;
	Xuint16 DeviceId = XPAR_ETHERNET_MAC_DEVICE_ID;


	/*
     * Initialize the XEmac component. The device ID chosen should be
     * configured without DMA (see the xemac_g.c file)
     */
    Status = XEmac_Initialize(InstancePtr, DeviceId);
    if (Status != XST_SUCCESS)
    {
        return XST_FAILURE;
    }

    /*
     * Run self-test on the device, which verifies basic sanity of the
     * device and the driver.
     */
    Status = XEmac_SelfTest(InstancePtr);
    if (Status != XST_SUCCESS)
    {
        return XST_FAILURE;
    }

    /*
     * Configure the device to be promiscuous
     */
    Options = XEmac_GetOptions(InstancePtr);
    Options |= (XEM_INSERT_FCS_OPTION |
				XEM_INSERT_PAD_OPTION |
				XEM_UNICAST_OPTION |
				XEM_BROADCAST_OPTION |
				XEM_POLLED_OPTION|
				XEM_PROMISC_OPTION|
				XEM_STRIP_PAD_FCS_OPTION);

    Status = XEmac_SetOptions(InstancePtr, Options);
    if (Status != XST_SUCCESS)
    {
        return XST_FAILURE;
    }

    /*
     * Set the MAC Addrees
     */
    Status = XEmac_SetMacAddress(InstancePtr, LocalAddress); //Note: Even though we are setting a MAC
															 //Address for the node, it is in promiscuous
															 //mode
    if (Status != XST_SUCCESS)
    {
        return XST_FAILURE;
    }

    /*
     * Start the device, which enables the transmitter and receiver
     */
    Status = XEmac_Start(InstancePtr);
    if (Status != XST_SUCCESS)
    {
        return XST_FAILURE;
    }

    Status = XEmac_Initialize(InstancePtr, DeviceId);
    if (Status != XST_SUCCESS)
    {
        return XST_FAILURE;
    }

    /*
     * Run self-test on the device, which verifies basic sanity of the
     * device and the driver.
     */
    Status = XEmac_SelfTest(InstancePtr);
    if (Status != XST_SUCCESS)
    {
        return XST_FAILURE;
    }
	xil_printf("complete!\r\n");

	xil_printf("	Allocating Xilkernel Buffer...");
	l = bufcreate (&mbuftACK, membufACK, BUFWIDTH, BUFHEIGHT);
  	if (l == -1){
        xil_printf ("ACKMAC: Error while creating memory pool. Errno: %d.\r\n", errno);
        return NULL;
    }
	xil_printf("complete!\r\n");
	///////////////////////////////////////////////////////////////////////////////////////////////////

	/////////////////////////////////Set up ACKMAC////////////////////////////////////////////////////
	pthread_attr_init (&attr);
	pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
	xil_printf("	Initializing controlStruct...");

	controlStruct.enableSequencing = 0;
	controlStruct.packetDupe=0;
	controlStruct.mySequenceNum.sequenceNumber[1] = 0;
	controlStruct.maxReSend = 4;
	controlStruct.currBackoff = 0;
	controlStruct.timerEnabled = 0;
	controlStruct.maxBackoff = 5;
	xil_printf("complete!\r\n");

	int retval;

	xil_printf("	Initializing Interrupts...");
	// Initialize the LEDs
	XGpio_mSetDataReg(XPAR_LEDS_4BIT_BASEADDR, 1, 0x0004);

	// Initialize the push buttons
	XGpio_Initialize(&Gpio, XPAR_PUSH_BUTTONS_4BIT_DEVICE_ID);

	XGpio_mSetDataDirection(XPAR_PUSH_BUTTONS_4BIT_BASEADDR, 1, 0x3);
	// Enable interrupts for the push buttons
	XGpio_InterruptEnable(&Gpio, XGPIO_IR_CH1_MASK);

	// Enable global interrupts. Both this and above line necessary to enable interrupts
	XGpio_InterruptGlobalEnable(&Gpio);

	XIntc_SetIntrSvcOption(XPAR_OPB_INTC_0_BASEADDR,XIN_SVC_ALL_ISRS_OPTION);

	// Start the interrupt controller
	XIntc_mMasterEnable(XPAR_OPB_INTC_0_BASEADDR);

	// Enable PPC non-critical interrupts
	XExc_mEnableExceptions(XEXC_NON_CRITICAL);

	register_int_handler(XPAR_OPB_INTC_0_PUSH_BUTTONS_4BIT_IP2INTC_IRPT_INTR, pb_int_handler, NULL);
	register_int_handler(XPAR_OPB_INTC_0_OFDM_TXRX_MIMO_OPBW_0_RX_INT_GOODPKT_INTR, rxPhyGood, NULL);
	register_int_handler(XPAR_OPB_INTC_0_OFDM_TXRX_MIMO_OPBW_0_RX_INT_BADPKT_INTR, rxPhyBad, NULL);
	register_int_handler(XPAR_OPB_INTC_0_OFDM_TIMER_OPBW_0_TIMEREXPIRE_INTR, timer_a_int_handler, NULL);

	enable_interrupt(XPAR_OPB_INTC_0_PUSH_BUTTONS_4BIT_IP2INTC_IRPT_INTR);
	enable_interrupt(XPAR_OPB_INTC_0_OFDM_TXRX_MIMO_OPBW_0_RX_INT_GOODPKT_INTR);
	enable_interrupt(XPAR_OPB_INTC_0_OFDM_TXRX_MIMO_OPBW_0_RX_INT_BADPKT_INTR);
	enable_interrupt(XPAR_OPB_INTC_0_OFDM_TIMER_OPBW_0_TIMEREXPIRE_INTR);
	xil_printf("complete!\r\n");

	////////////////////////////////////SET UP RADIO and PHY///////////////////////////////////////////

	xil_printf("	Initializing Radio Transmitter...");
	//Reset the radio transceiver's registers
	//Setup the SPI controller to control the DAC
	WarpDac_Reset(XPAR_RADIO_CONTROLLER_0_BASEADDR, FIRST_DAC | SECOND_DAC);
	WarpDac_InitializeSPI(XPAR_RADIO_CONTROLLER_0_BASEADDR);
	WarpDac_InitializeDAC(XPAR_RADIO_CONTROLLER_0_BASEADDR, FIRST_DAC | SECOND_DAC);

	warpmac_CalibrateTxDCO();

	//Setup the radio controller & reset radio #2
	WarpRadio_Reset(XPAR_RADIO_CONTROLLER_0_BASEADDR, FIRST_RADIO | SECOND_RADIO);
	WarpRadio_InitializeSPI(XPAR_RADIO_CONTROLLER_0_BASEADDR);
	WarpRadio_InitializeRadio(XPAR_RADIO_CONTROLLER_0_BASEADDR, FIRST_RADIO | SECOND_RADIO);

	//Enable SPI control of Tx gain
	WarpRadio_SerialTxGain(XPAR_RADIO_CONTROLLER_0_BASEADDR, 0, FIRST_RADIO | SECOND_RADIO);

	//Set Tx gain (6-bit value- 0x00 min, 0x3f max)

	WarpRadio_TxVGAGainControl(XPAR_RADIO_CONTROLLER_0_BASEADDR, 0x3f, FIRST_RADIO | SECOND_RADIO); //3f

	//Set Tx LPF to +/-12MHz
	WarpRadio_TxLpfCornFreqCoarseAdj(XPAR_RADIO_CONTROLLER_0_BASEADDR, 0x1, FIRST_RADIO | SECOND_RADIO);

	//Set Tx baseband gain min
	WarpRadio_BaseBandTxGain(XPAR_RADIO_CONTROLLER_0_BASEADDR, 0x3, FIRST_RADIO | SECOND_RADIO);

	//WarpRadio_SetCenterFreq2GHz(XPAR_RADIO_CONTROLLER_0_BASEADDR, 11, FIRST_RADIO | SECOND_RADIO);
	usleep(100);


	//Finally turn on the power amplifier
	WarpRadio_24AmpEnable(XPAR_RADIO_CONTROLLER_0_BASEADDR, FIRST_RADIO | SECOND_RADIO);
	WarpRadio_SetTxTiming(XPAR_RADIO_CONTROLLER_0_BASEADDR, FIRST_RADIO, 50, 0, 0, 0); //radio 2
	WarpRadio_SetTxTiming(XPAR_RADIO_CONTROLLER_0_BASEADDR, SECOND_RADIO, 76+50, 76, 76, 76); //radio 1
	WarpRadio_SetTxGainTiming(XPAR_RADIO_CONTROLLER_0_BASEADDR, FIRST_RADIO | SECOND_RADIO, 0x3F, 0xF, 2); //3f
	WarpRadio_SetMimoMode(XPAR_RADIO_CONTROLLER_0_BASEADDR, 1, FIRST_RADIO | SECOND_RADIO);

	xil_printf("complete!\r\n");
	//**** End Radio Transmitter Initialization ****

	//**** Begin Radio Receiver Initialization ****

	xil_printf("	Initializing Radio Receiver...");
	WarpRadio_SerialRxGain(XPAR_RADIO_CONTROLLER_0_BASEADDR, 0, FIRST_RADIO | SECOND_RADIO); //Allows AGC to control gains
																							 //	WarpRadio_SerialRxGain(XPAR_RADIO_CONTROLLER_0_BASEADDR, 1, FIRST_RADIO | SECOND_RADIO); //Allows AGC to control gains

	usleep(10);
	WarpRadio_RxLpfCornFreqCoarseAdj(XPAR_RADIO_CONTROLLER_0_BASEADDR, 0x1, FIRST_RADIO | SECOND_RADIO);
	usleep(10);

	xil_printf("complete\r\n");
	//**** End Radio Receiver Initialization ****


	//**** Begin OFDM Transmitter Initialization ****

	xil_printf("	Initializing OFDM Transmitter...");

#define TX_FFT_SCALING_STAGE1 2 //was 1, which sucked for VLC, inexplicably
#define TX_FFT_SCALING_STAGE2 2
#define TX_FFT_SCALING_STAGE3 3

	mimo_ofdmTx_init();
	mimo_ofdmTx_setFFTScaling((unsigned int)(16*TX_FFT_SCALING_STAGE1 + 4*TX_FFT_SCALING_STAGE2 + 1*TX_FFT_SCALING_STAGE3));
	mimo_ofdmTx_setPilot1Index((7 + ((64-7)<<16) ));
	mimo_ofdmTx_setPilot1Value(0xc000);//c000 8000
		mimo_ofdmTx_setPilot2Index((21 + ((64-21)<<16)));
		mimo_ofdmTx_setPilot2Value(0x3fff);  //3fff 7fff
		mimo_ofdmTx_setRandom(0);

		xil_printf("numTxBytes: %d\r\n",ofdm_TxRx_mimo_ReadReg_Tx_NumPayloadBytes(XPAR_OFDM_TXRX_MIMO_OPBW_0_BASEADDR));

		mimo_ofdmTx_setNumSyms(numTrainingSyms + (numBaseRate*256) + (40*65536));
		mimo_ofdmTx_setPreambleScaling(0x40FF);//0x40FF); //7fff
			mimo_ofdmTx_setControlBits(
									   (4 << 4) |
									   (TX_SISO_MODE * sisoMode)|
									   /*TX_RANDOM_MODE |*/
									   /*TX_DISABLE_ANTB_PREAMBLE |*/
									   TX_PILOT_SCRAMBLING
									   );
			mimo_ofdmTx_enable();

			//**** End OFDM Transmitter Initialization ****

			//**** Begin OFDM Receiver Initialization ****

			xil_printf("	Initializing OFDM Receiver...");
#define FFT_SCALING_STAGE1 0	//1
#define FFT_SCALING_STAGE2 1	//2
#define FFT_SCALING_STAGE3 1	//1

			mimo_ofdmRx_init();

			//Scaling options:
			//268439552 -> 2
			//201329664 -> 1.5
			//134219776 -> 1
			unsigned int myScaling = 201329664;//(2048 + 2048<<16);

				mimo_ofdmRx_setFFTScaling((unsigned int)(16*FFT_SCALING_STAGE1 + 4*FFT_SCALING_STAGE2 + 1*FFT_SCALING_STAGE3));
				mimo_ofdmRx_setFFTWindowOffset(14); //8
				mimo_ofdmRx_setRxScaling(myScaling);//134219776);//(unsigned int)((unsigned int)2048 + ((unsigned int)2048*(2^16))));//2048 + 2048<<16);//2048 + 2048*2^16);//134219776);//0x1e66 + (0x1e66<<16)); //.5 in Fix_16_11
					mimo_ofdmRx_setCorrThresh(0x60);
					mimo_ofdmRx_setEnergyThresh(0x2);
					mimo_ofdmRx_setLongCorrParams(251-16  + (4000*65536)); //251 2500
					mimo_ofdmRx_setPktDetDly(40); //30 //+43, //52
					mimo_ofdmRx_setNumOFDMSyms(numTrainingSyms + (numBaseRate*65536));

					mimo_ofdmRx_setByteNums( (unsigned int)(24 + (16<<16) + (17<<24) ));

					mimo_ofdmRx_setCFOKI(INIT_KIVAL);
					mimo_ofdmRx_setCFOKP(INIT_KPVAL);

					mimo_ofdmRx_setOptions( /*RESET_BER |*/
						REQ_LONG_CORR |
						DYNAMC_PKT_LENGTHS |
						//DYNAMIC_MOD |
						REQ_TWO_LONG_CORR |
						REQ_SHORT_CORR |
						(RX_SISO_MODE * sisoMode) |
						//CFO_USE_STS |
						CFO_USE_LTS |
						CFO_USE_LONGCORR |
						CFO_USE_ARCTAN |
						//BYPASS_CARR_REC |
						EXT_PKT_DETECT,
						//INT_PKT_DETECT,
						//EQ_BYPASS_DIVISION,
						INTR_BAD_PKTS | INTR_GOOD_PKTS
						);
mimo_ofdmRx_enable();
unsigned int radio = FIRST_RADIO;
if(sisoMode==0) radio = radio | SECOND_RADIO;
WarpRadio_RxEnable(XPAR_RADIO_CONTROLLER_0_BASEADDR, radio);
usleep(100);
usleep(100);
xil_printf("complete!\r\n");
//**** End OFDM Receiver Initialization ****

xil_printf("	Initializing AGC...");

///////////////////////////////SET UP AGC///////////////////////
//Enable HARDWARE control of the RxHP pin
WarpRadio_RxHp_Soft_Control_Disable(XPAR_RADIO_CONTROLLER_0_BASEADDR, FIRST_RADIO | SECOND_RADIO);
WarpRadio_RxHighPassCornerFreq(XPAR_RADIO_CONTROLLER_0_BASEADDR, 0, FIRST_RADIO | SECOND_RADIO);

ofdm_AGC_Initialize(agcnoiseEst);
ofdm_AGC_setNoiseEstimate(agcnoiseEst);
ofdm_AGC_FiltSel(1);
ofdm_AGC_SetDCO(1); //1
ofdm_AGC_SetTarget(agctarget); //-7
ofdm_AGC_Reset();
xil_printf("complete!\r\n");
////////////////////////////////////////////////////////////////
////////////////SET UP RSSI PACKET DETECTOR/////////////////////
xil_printf("	Initializing Packet Detection...");
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_csma_enableBusy_OFFSET)) =1;
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_csma_enableIdle_OFFSET)) =1;
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_masterReset_OFFSET)) = 1;
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_detectionMask_OFFSET)) = 2; //3= Radios 1 and 2 **needs help**
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_detectionMode_OFFSET)) = 1; //0 = AND, 1 = OR
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_resetDuration_OFFSET)) = 32;
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_avgLen_OFFSET)) = 16;
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_avgThresh_OFFSET)) = 7000; //7000
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_csma_avgThresh_OFFSET)) = 5000;//7000
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_csma_difsPeriod_OFFSET)) = 625;
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_masterReset_OFFSET)) = 0;
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_reset_OFFSET)) = 0;
xil_printf("complete!\r\n");
////////////////////////////////////////////////////////////////
warpmac_setNumTrainingSyms(4);
}


void warpmac_pollEthernet(int* callback()){

	Xuint32 FrameLen = MY_XEM_MAX_FRAME_SIZE;
	XStatus Status;
	Xuint32 Options;

	FrameLen = MY_XEM_MAX_FRAME_SIZE;

	Status = XEmac_PollRecv(InstancePtr, (Xuint8 *)RXBuffer, &FrameLen);

	if (Status != XST_SUCCESS)
	{
		if (!(Status == XST_NO_DATA || Status == XST_BUFFER_TOO_SMALL))
		{
			XEmac_Reset(InstancePtr);
			XEmac_SetMacAddress(InstancePtr,
								LocalAddress);
			Options = ( XEM_INSERT_FCS_OPTION |
						XEM_INSERT_PAD_OPTION |
						XEM_UNICAST_OPTION |
						XEM_BROADCAST_OPTION |
						XEM_POLLED_OPTION|
						XEM_PROMISC_OPTION|
						XEM_STRIP_PAD_FCS_OPTION
						);
			XEmac_SetOptions(InstancePtr, Options);
			XEmac_Start(InstancePtr);
		}
	}
	else{

		*(int*)callback((Xuint8 *)RXBuffer, FrameLen);

	}



}

/*WARPMAC_SENDETHERNET- Pushes the given Macframe over ethernet*/
void warpmac_sendEthernet(Macframe* packet){


	if(!((controlStruct.enableSequencing) && (controlStruct.packetDupe))){
		unsigned int length;
		length = (packet->length)-4;
		XEmac_PollSend(InstancePtr, packet->data, length);
	}

}


/*WARPMAC_SENDOFDM- performs the conversion from the packet structure to the byte array and
sends the packet over the air*/
int warpmac_sendOfdm(Macframe* packet){
    int i;
    int numPayloadBytes;
    int dataStartIndex;
    int Npld;

	unsigned int radio = FIRST_RADIO;
	if(sisoMode==0) radio = radio | SECOND_RADIO;

	*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_reset_OFFSET)) = 1;
	//Disable the receiver
	WarpRadio_RxDisable(XPAR_RADIO_CONTROLLER_0_BASEADDR, radio);


	Npld = 2 + ((packet->length + 4 + (6*RATE-1))/(6*RATE)) * (sisoMode+1);
	//		Npld = 2 + ((packet->length + 4 + (6*RATE-1))/(6*RATE));
	//		if(sisoMode) Npld = Npld<<1;
	//		Npld = ((packet->length + 4 + (6*RATE-1))/(6*RATE))*(sisoMode+1);

	numPayloadBytes = 24 + EXTRASYMBOLS*12 + packet->length + 4;

	mimo_ofdmTx_setNumBytes(numPayloadBytes);
	mimo_ofdmTx_setNumSyms(numTrainingSyms + (numBaseRate*256) + (Npld*65536));
	unsigned char header[24];
	dataStartIndex = 24+12*EXTRASYMBOLS;

	header[0]= (packet->destAddr[3]);
	header[1]= (packet->destAddr[2]);
	header[2]= (packet->destAddr[1]);
	header[3]= (packet->destAddr[0]);
	header[4]= (packet->srcAddr[1]);
	header[5]= (packet->srcAddr[0]);
	header[6]= (packet->destAddr[5]);
	header[7]= (packet->destAddr[4]);
	header[8]= (packet->srcAddr[5]);
	header[9]= (packet->srcAddr[4]);
	header[10]= (packet->srcAddr[3]);
	header[11]= (packet->srcAddr[2]);
	header[12]= Npld;
	header[13]= packet->pktType;
	header[14]= RATE;
	header[15]= packet->pktRev;  //version/mode/revision
	header[16]= (controlStruct.mySequenceNum.sequenceNumber[1]); //0
	header[17]= 0; //0
	header[18]= numPayloadBytes >> 8;              //most significant bits of total length*/
	header[19]= numPayloadBytes;                //least significant bits of total length
	header[20]= 0;
	header[21]= 0;
	header[22]= 0;
	header[23]= 0;



	mimo_ofdmTx_copyHeader(header,(size_t)24);
	mimo_ofdmTx_copyPayload(packet->data,packet->length);


	mimo_mac2phy_pktTx(TXBLOCK,radio);

	WarpRadio_RxEnable(XPAR_RADIO_CONTROLLER_0_BASEADDR, radio);
	WarpRadio_RxHp_Soft_Control_Disable(XPAR_RADIO_CONTROLLER_0_BASEADDR, FIRST_RADIO | SECOND_RADIO);

	if(globalReset==0) *((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_reset_OFFSET)) = 0;
	return 0;
}

/*WARPMAC_SETUPBUTTONHANDLER- attaches user callback to up button*/
void warpmac_setUpButtonHandler(void(*handler)()){
	usr_upbutton = handler;
}

/*WARPMAC_SETLEFTBUTTONHANDLER- attaches user callback to left button*/
void warpmac_setLeftButtonHandler(void(*handler)()){
	usr_leftbutton = handler;
}

/*WARPMAC_SETRIGHTBUTTONHANDLER- attaches user callback to right button*/
void warpmac_setRightButtonHandler(void(*handler)()){
	usr_rightbutton = handler;
}

/*WARPMAC_SETMIDDLEBUTTONHANDLER- attaches user callback to middle button*/
void warpmac_setMiddleButtonHandler(void(*handler)()){
	usr_middlebutton = handler;
}

/*WARPMAC_SETTIMERHANDLER- attaches user callback to timer*/
void warpmac_setTimerHandler(void(*handler)()){
	usr_timerHandler = handler;
}

/*WARPMAC_SETGOODPACKETHANDLER- attaches user callback to good packet interrupt*/
void warpmac_setGoodPacketHandler(void(*handler)()){
	usr_goodPacketHandler = handler;
}

/*WARPMAC_SETBADPACKETHANDLER- attaches user callback to bad packet interrupt*/
void warpmac_setBadPacketHandler(void(*handler)()){
	usr_badPacketHandler = handler;
}


/*WARPMAC_INCREMENTRESEND- increments the resend counter for the current transmit.
Additionally, it will free the payload when the maximum number of retransmissions has
occured.*/
int warpmac_incrementResend(Macframe* packet){
	int status = 1;
	packet->currReSend = ( packet->currReSend+1)%((controlStruct.maxReSend)+1); //increment currReSend and wrap around max
	if( packet->currReSend == 0) { //if reSend wraps around, the packet will be dropped
		status = 0;
		buffree(mbuftACK,packet->data);
		controlStruct.currBackoff=0;
	   	packet->isNew = 0;
	}
	return status;
}


/*WARPMAC_ALLOCATEPAYLOAD- readies a Macframe for transmission. This includes
allocating memory for the payload and setting Macframe parameters.*/
int warpmac_allocatePayload(Macframe* packet,unsigned int length){
	/*Function tells the microkernel to allocate space for the packet payload*/
	packet->data = (unsigned char*)bufmalloc(mbuftACK,sizeof(char)*length);
	if(packet->data==NULL) xil_printf("DATASOURCE: Error allocating .data  Errno: %d.\r\n", errno);
	//packet->currReSend = 0;
	//packet->isNew = 1;
	return packet->data; //if NULL, allocation failed
}

/*WARPMAC_FREEPAYLOAD- Frees previously allocated memory. Proper use of this
function ensures no memory leaks due to unallocated memory.*/
int warpmac_freePayload(Macframe* packet){
	/*Function tells the microkernel to allocate space for the packet payload*/
	(controlStruct.mySequenceNum.sequenceNumber[1])++;
	//buffree(mbuftACK,packet->data);
	packet->currReSend = 0;
	packet->isNew = 0;
	return 0;
}

/*WARPMAC_ADDRESSEDTOME- returns a value corresponding to whether the packet was destined
for this particular node.*/
int warpmac_addressedToMe(Macframe* packet){
	int i,sum;
	unsigned char tempBoolean;
	sum = 0;
	for(i=0;i<6;i++){
		sum = sum+(packet->destAddr[i] == controlStruct.selfAddr[i]);
				}
				tempBoolean = 0;
				if(sum==6) tempBoolean = 1;
				sum = 0;
				return tempBoolean;
}

/*WARPMAC_ADDRESSEDFROMTHEM- creturns a value corresponding to whether the packet was sourced
from a given address*/
int warpmac_addressedFromThem(Macframe* packet,unsigned char* addr){
	int i,sum;
	unsigned char tempBoolean;
	sum = 0;
	for(i=0;i<6;i++){
		sum = sum+(packet->srcAddr[i] == addr[i]);
		//xil_printf("%x == ",packet->srcAddr[i]);
		//xil_printf("%x\r\n",addr[i]);
				}
				tempBoolean = 0;
				if(sum==6) tempBoolean = 1;
				sum = 0;

				return tempBoolean;
}


/*WARPMAC_ADDRESSEDTOTHEM- creturns a value corresponding to whether the packet was sourced
from a given address*/
int warpmac_addressedToThem(Macframe* packet,unsigned char* addr){
	int i,sum;
	unsigned char tempBoolean;
	sum = 0;
	for(i=0;i<6;i++){
		sum = sum+(packet->destAddr[i] == addr[i]);
				}
				tempBoolean = 0;
				if(sum==6) tempBoolean = 1;
				sum = 0;

				return tempBoolean;
}




/*WARPMAC_SETMAXRESEND- sets the maximum number of resends. This value is used by
WARPMAC_INCREMENTRESEND.*/
void warpmac_setMaxResend(unsigned int c){
	controlStruct.maxReSend=c;
}

/*WARPMAC_SETMAXCW- sets the maximum contention window.*/
void warpmac_setMaxCW(unsigned int c){
	controlStruct.maxBackoff=c;
}

/*WARPMAC_SETCHANNEL- sets the frequency band to operate in.*/
void warpmac_setChannel(unsigned int c){
unsigned int val;
	val = WarpRadio_SetCenterFreq2GHz(XPAR_RADIO_CONTROLLER_0_BASEADDR, c, FIRST_RADIO | SECOND_RADIO);

}

/*WARPMAC_SETMACADDR- sets wireless MAC address of this node.*/
int warpmac_setMacAddr(unsigned char* addr){
	controlStruct.selfAddr[5] = addr[5];
	controlStruct.selfAddr[4] = addr[4];
	controlStruct.selfAddr[3] = addr[3];
	controlStruct.selfAddr[2] = addr[2];
	controlStruct.selfAddr[1] = addr[1];
	controlStruct.selfAddr[0] = addr[0];
}

/*WARPMAC_SETTIMEOUT- sets the amount of time the node is willing to wait for an acknowledgement.*/
int warpmac_setTimeout(unsigned int time){
	controlStruct.timeout = time;
}

/*WARPMAC_SETSLOTTIME- sets the base backoff window.*/
int warpmac_setSlotTime(unsigned int time){
	controlStruct.slotTime = time;
}

/*WARPMAC_INTIMEOUT- returns a value corresponding to whether or not the node is in timeout.*/
int warpmac_inTimeout(){
	return controlStruct.timerType==TIMEOUT;
}

/*WARPMAC_GETMYID- reads the value from the user dip switches for use as node identification.*/
int warpmac_getMyId(){
	int ret;
	ret =  XGpio_mGetDataReg(XPAR_DIPSWS_4BIT_BASEADDR,1);
	XGpio_mSetDataReg(XPAR_LED_7SEGMENT_BASEADDR, 1, sevenSegmentMap(ret));
	//This function returns the value currently specified in the Dip Switch on the FPGA board.
	return ret;
}

/*WARPMAC_COPYPAYLOAD- copies the payload from the received to the provided Macframe.*/
int warpmac_copyPayload(Macframe* packet){

	if(!((controlStruct.enableSequencing) && (controlStruct.packetDupe))){

		mimo_ofdmRx_copyPayload(packet->data,packet->length);

	}

}


void warpmac_CalibrateTxDCO () {
	unsigned char i;
	int eepromStatus = 0;
	short calReadback = 0;
	signed short best_I, best_Q;

	for(i=1; i<(3); i++)
	{
		//EEPROM Stuff
		eepromStatus = WarpEEPROM_EEPROMSelect((unsigned int*)XPAR_EEPROM_0_BASEADDR, i);
		eepromStatus = WarpEEPROM_Initialize((unsigned int*)XPAR_EEPROM_0_BASEADDR);
		calReadback = WarpEEPROM_ReadRadioCal((unsigned int*)XPAR_EEPROM_0_BASEADDR, 2, 0);

		best_I = (signed short)(((signed char)(calReadback & 0xFF))<<1);
		best_Q = (signed short)(((signed char)((calReadback>>8) & 0xFF))<<1);

		xil_printf("I: %d Q: %d\r\n", best_I, best_Q);

		//Switch the SPI controller to DAC mode
		WarpDac_InitializeSPI(XPAR_RADIO_CONTROLLER_0_BASEADDR);

		//Initialize the DAC
		WarpDac_InitializeDAC(XPAR_RADIO_CONTROLLER_0_BASEADDR, 0x1<<(3+i));

		WarpDac_OffsetAdj(XPAR_RADIO_CONTROLLER_0_BASEADDR,ICHAN,best_I,0x1<<(3+i));
		WarpDac_OffsetAdj(XPAR_RADIO_CONTROLLER_0_BASEADDR,QCHAN,best_Q,0x1<<(3+i));
	}

	return;
}

/*WARPMAC_ENABLESISOMODE- sets the MIMO OFDM core into a SISO mode for legacy reasons.*/
void warpmac_enableSisoMode(){
	sisoMode = 1;

	mimo_ofdmTx_setControlBits(
							   (4 << 4) |
							   (TX_SISO_MODE * sisoMode) |
							   /*TX_RANDOM_MODE |*/
							   /*TX_DISABLE_ANTB_PREAMBLE |*/
							   TX_PILOT_SCRAMBLING
							   );

	mimo_ofdmRx_setOptions(
						   RESET_BER |
						   REQ_LONG_CORR |
						   DYNAMC_PKT_LENGTHS |
						   REQ_TWO_LONG_CORR |
						   REQ_SHORT_CORR |
						   (RX_SISO_MODE * sisoMode) |
						   //CFO_USE_STS |
						   CFO_USE_LTS |
						   CFO_USE_LONGCORR |
						   CFO_USE_ARCTAN |
						   //BYPASS_CARR_REC |
						   EXT_PKT_DETECT |
						   CFO_USE_TRAINING,
						   //EQ_BYPASS_DIVISION,
						   INTR_BAD_PKTS | INTR_GOOD_PKTS
						   );
}

/*WARPMAC_ENABLEMIMOMODE- sets the MIMO OFDM core into the MIMO mode.*/
void warpmac_enableMimoMode(){
	sisoMode = 0;
	mimo_ofdmTx_setControlBits(
							   (4 << 4) |
							   (TX_SISO_MODE * sisoMode) |
							   /*TX_RANDOM_MODE |*/
							   /*TX_DISABLE_ANTB_PREAMBLE |*/
							   TX_PILOT_SCRAMBLING
							   );

	mimo_ofdmRx_setOptions(
						   RESET_BER |
						   REQ_LONG_CORR |
						   DYNAMC_PKT_LENGTHS |
						   REQ_TWO_LONG_CORR |
						   REQ_SHORT_CORR |
						   (RX_SISO_MODE * sisoMode) |
						   //CFO_USE_STS |
						   CFO_USE_LTS |
						   CFO_USE_LONGCORR |
						   CFO_USE_ARCTAN |
						   //BYPASS_CARR_REC |
						   EXT_PKT_DETECT,
						   //EQ_BYPASS_DIVISION,
						   INTR_BAD_PKTS | INTR_GOOD_PKTS
						   );
}

/*WARPMAC_SETNUMTRAININGSYMS- sets the number of training symbols
used by the OFDM core*/
void warpmac_setNumTrainingSyms(unsigned int c){
	numTrainingSyms=c; //4
	mimo_ofdmRx_setNumOFDMSyms(numTrainingSyms + (numBaseRate*65536));
}

unsigned char sevenSegmentMap(unsigned char x)
{
	switch(x)
	{
		case(0x0) : return 0x007E;
		case(0x1) : return 0x0030;
		case(0x2) : return 0x006D;
		case(0x3) : return 0x0079;
		case(0x4) : return 0x0033;
		case(0x5) : return 0x005B;
		case(0x6) : return 0x005F;
		case(0x7) : return 0x0070;
		case(0x8) : return 0x007F;
		case(0x9) : return 0x007B;

		case(0xA) : return 0x0077;
		case(0xB) : return 0x007F;
		case(0xC) : return 0x004E;
		case(0xD) : return 0x007E;
		case(0xE) : return 0x004F;
		case(0xF) : return 0x0047;
default : return 0x0000;
	}
}

void warpmac_enableCSMA(){
	*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_csma_avgThresh_OFFSET)) = 5000;//3000
}

void warpmac_disableCSMA(){
	*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_csma_avgThresh_OFFSET)) = 16000;//16000
}

/*WARPMAC_ENABLESEQUENCING- prevents duplicate packets from appearing over the wire*/
void warpmac_enableSequencing(){
	controlStruct.enableSequencing=1;
}

/*WARPMAC_DISABLESEQUENCING- allows duplicate packets from appearing over the wire*/
void warpmac_disableSequencing(){
	controlStruct.enableSequencing=0;
}


void warpmac_setPktDlyPlus(){
	pktDly++;
	xil_printf("+PktDly = %d\r\n",pktDly);
	mimo_ofdmRx_setPktDetDly(pktDly);
}
void warpmac_setPktDlyMinus(){
	pktDly--;
	xil_printf("-PktDly = %d\r\n",pktDly);
	mimo_ofdmRx_setPktDetDly(pktDly);
}

void warpmac_setKPPlus(unsigned int increment){
	KPval=KPval+increment;
	//xil_printf("+KPval = %x\r\n",KPval);
	mimo_ofdmRx_setCFOKP(KPval);
	xil_printf("I: %x\tP: %x\r\n", KIval, KPval);
}

void warpmac_setKPMinus(unsigned int decrement){
	KPval=KPval-decrement;
	//xil_printf("-KPval = %x\r\n",KPval);
	mimo_ofdmRx_setCFOKP(KPval);
	xil_printf("I: %x\tP: %x\r\n", KIval, KPval);
}

void warpmac_setKIPlus(unsigned int increment){
	KIval=KIval+increment;
	//xil_printf("+KIval = %x\r\n",KIval);
	mimo_ofdmRx_setCFOKI(KIval);
	xil_printf("I: %x\tP: %x\r\n", KIval, KPval);
}


void warpmac_setKIMinus(unsigned int decrement){
	KIval=KIval-decrement;
	//xil_printf("-KIval = %x\r\n",KIval);
	mimo_ofdmRx_setCFOKI(KIval);
	xil_printf("I: %x\tP: %x\r\n", KIval, KPval);
}



void warpmac_setNoiseTargetPlus(){
	agcnoiseEst++;
	xil_printf("+Noise Estimate = %d\r\n",agcnoiseEst);
	ofdm_AGC_setNoiseEstimate(agcnoiseEst);
}

void warpmac_setNoiseTargetMinus(){
	agcnoiseEst--;
	xil_printf("-Noise Estimate = %d\r\n",agcnoiseEst);
	ofdm_AGC_setNoiseEstimate(agcnoiseEst);
}

void warpmac_setTargetPlus(){
	agctarget++;
	xil_printf("+agctarget = %d\r\n",agctarget);
	XGpio_mSetDataReg(XPAR_LED_7SEGMENT_BASEADDR, 1, sevenSegmentMap(-agctarget%10));
	XGpio_mSetDataReg(XPAR_LED_7SEGMENT_1_BASEADDR, 1, sevenSegmentMap(-agctarget/10));
	ofdm_AGC_SetTarget(agctarget);
}

void warpmac_setTargetMinus(){
	agctarget=agctarget-1;
	xil_printf("-agctarget = %d\r\n",agctarget);
	XGpio_mSetDataReg(XPAR_LED_7SEGMENT_BASEADDR, 1, sevenSegmentMap(-agctarget%10));
	XGpio_mSetDataReg(XPAR_LED_7SEGMENT_1_BASEADDR, 1, sevenSegmentMap(-agctarget/10));
	ofdm_AGC_SetTarget(agctarget);
}

void warpmac_leftHex(unsigned char x){
	XGpio_mSetDataReg(XPAR_LED_7SEGMENT_1_BASEADDR, 1, sevenSegmentMap(x));
}
void warpmac_rightHex(unsigned char x){
	XGpio_mSetDataReg(XPAR_LED_7SEGMENT_BASEADDR, 1, sevenSegmentMap(x));
}

void warpmac_disableDetection(){
globalReset = 1;
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_reset_OFFSET)) = 1;

}

void warpmac_enableDetection(){
globalReset = 0;
*((volatile unsigned int *)(XPAR_OFDM_PKTDETECTOR_MIMO_OPBW_0_BASEADDR+ofdm_pktDetector_mimo_pktDet_reset_OFFSET)) = 0;
}

void warpmac_setRxBuffer(Macframe* packet){
	rxPacket=packet;
}

void warpmac_printDestAddr(Macframe* packet){
	int i,sum;
	unsigned char tempBoolean;
	sum = 0;
	for(i=0;i<6;i++){
		xil_printf("%x\r\n",packet->destAddr[i]);
				}

}

